/*-------------------------------------------------------------------------------
Sequential Algorithm for Maximum Flow in Flow Network using OpenMP
Flow Graph represented as Matrix - G[i][j] => flow between ith and jth vertex in graph G  
Author: V.Sai Venkata Krishnan
Input format: ./ProgName
Assuming vertices are numbered 0 to V-1
---------------------------------------------------------------------------------*/

#include<iostream>
#include<vector>
#include<queue>
#include<tuple>
#include<climits>
#include<ctime>
#include<fstream>

using namespace std;

long s;
long d;

void readGraph2(vector< vector<int> > &graph)
{	string graphName;
	cout<<"\nEnter the name of graph with extension\n";
	cin>>graphName;
	ifstream myfile1(graphName,ios_base::in);
	
	long a;
	long b;
	int flow;
	long n;
	long m;
	long i;
	long j;

	cout<<"\nEnter the no of nodes\n";
	cin>>n;
	
	graph.resize(n); 

	for(i=0;i<n;i++)
	{	graph[i].resize(n);
		for(j=0;j<n;j++)
		{	graph[i][j]=0;
				
		}
	}
	while(myfile1>>a)
	{	myfile1>>b;
		myfile1>>flow;
		graph[a][b]=flow;
	}	
	//cout<<graph.size()<<endl;	
	
}



void displayGraph(vector< vector<int> > graph)
{	int i;
	int j;
	for(i=0;i<graph.size();i++)
	{	
		for(j=0;j<graph.size();j++)
		{	cout<<" "<<graph[i][j]<<" ";	
		}
		cout<<"\n";
	}
}
	
bool parallelBFS(vector< vector<int> > &graph,long* parent)
{	int i;	
	int j;
	int v;
	int u;
	bool* visited = new bool [graph.size()];
	queue<int> q;
	
	for(i=0;i<graph.size();i++)
	{	visited[i]=0;
	}

	visited[s] = true;

	q.push(s);
	parent[s]=-1;

	//standard bfs loop is parallelised using openmp constructs
	while(!q.empty())
	{	
		for(i=0;i<q.size();i++)
		{	
			{	u = q.front();
				q.pop();
			}
			for( j = 0;j<graph[u].size();j++)
			{	
				{			
					
					if(visited[j]==0 && graph[u][j] > 0)
					{	parent[j]=u;
						visited[j]=true;
						q.push(j); 
					
					}
	
				
				}
			}
		}
	}	
	return visited[d]==true ;
}



int fordFulkerson(vector< vector<int> > &graph)
{	long* parent = new long[graph.size()];
	int max_flow = 0;
	long v ;
	long u ;

	while(parallelBFS(graph,parent))
	{	
		int path_flow = INT_MAX;
				
		for(v=d ; v!=s ; v = parent[v])
		{	u = parent[v];			
			path_flow = min(path_flow,graph[u][v]);
		}
			
		for(v=d ; v!=s ; v=parent[v])
		{	u = parent[v];	
			graph[u][v] -= path_flow;
			graph[v][u] += path_flow;
		}
		max_flow += path_flow;
	}
	delete[] parent;
	return max_flow;

}

int main()
{	vector< vector<int> > graph;
	readGraph2(graph);
	cout<<"\nEnter source :"<<endl;
	cin>>s;
	cout<<"\nEnter destination :"<<endl;
	cin>>d;	
	int maxflow=0;
	clock_t t;
	t = clock();	
	maxflow = fordFulkerson(graph);
	t = clock() -t ;
	double timeTaken = ((double)t)/CLOCKS_PER_SEC;

	cout<<"\nMaximum flow is "<<maxflow<<endl;
	cout<<"\nTime taken = "<<timeTaken;			

	return 0;
}